@@ -1190,7 +1190,7 @@ SennaRemove(const RelFileNode *rnode)
 }
 
 static sen_query *
-SennaQuery(const char *sql, size_t len)
+SennaQuery(const char *str, size_t len)
 {
 	int					i;
 	SennaQueryCache    *cache;
@@ -1204,7 +1204,7 @@ SennaQuery(const char *sql, size_t len)
 
 		if (cache->query == NULL)
 			break;
-		if (strncmp(cache->key, sql, len) == 0)
+		if (strncmp(cache->key, str, len) == 0)
 			return cache->query;
 	}
 
@@ -1214,8 +1214,8 @@ SennaQuery(const char *sql, size_t len)
 		ereport(ERROR,
 				(errcode(ERRCODE_OUT_OF_MEMORY),
 				 errmsg("out of memory")));
-	strlcpy(key, sql, len + 1);
-	query = sen_query_open(sql, len, sen_sel_or, SENNA_MAX_N_EXPR, SennaEncoding());
+	strlcpy(key, str, len + 1);
+	query = sen_query_open(str, len, sen_sel_or, SENNA_MAX_N_EXPR, SennaEncoding());
 	if (query == NULL)
 		elog(ERROR, "senna: sen_query_open()");
 
@@ -1248,6 +1248,13 @@ SennaRecordsOpen(Relation indexRelation, const char *str, int len, sen_index **i
 	ListCell		   *cell;
 	SennaRecordsCache	*cache;
 
+	if (len == 0)
+		ereport(ERROR,
+			(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+			 errmsg("scan key must not be empty for senna index: %s.%s",
+					get_namespace_name(RelationGetNamespace(indexRelation)),
+					 RelationGetRelationName(indexRelation))));
+
 	query = SennaQuery(str, len);
 
 	/* search records cache first */
@@ -1261,6 +1268,7 @@ SennaRecordsOpen(Relation indexRelation, const char *str, int len, sen_index **i
 			records = cache->records;
 			if (idx)
 				*idx = SennaIndexOpen(indexRelation, 0);
+			sen_records_rewind(records);
 			return records;
 		}
 	}
@@ -1345,9 +1353,14 @@ SennaUnlock(Relation index, LOCKMODE lockmode)
 						 lockmode);
 }
 
-#define IS_LIKE_WILDCARD(c)		((c) == '%' || (c) == '_')
-#define LIKE_PRAGMA				"*D+ "
-#define LIKE_PRAGMA_LEN			4
+/*
+ * Note that senna cannot search " in words because the double quote is a
+ * special marker to separate words. I cannot find any descriptions in
+ * senna documentation to escape the character.
+ */
+#define LIKE_WILDCARD			"%_"
+#define WORD_SEP_CHARS			LIKE_WILDCARD "\""
+#define NON_SIMPLE_CHARS		WORD_SEP_CHARS " \\\t\v\r\n"
 
 /*
  * is_simple_like - true iff str matches %foo%.
@@ -1359,90 +1372,135 @@ is_simple_like(const char *str, int len)
 {
 	int		i;
 
-	if (len < 2 || str[0] != '%' || str[len - 1] != '%')
+	if (len <= 2 || str[0] != '%' || str[len - 1] != '%')
 		return false;
 
 	for (i = 1; i < len - 1; i++)
 	{
-		if (IS_LIKE_WILDCARD(str[i]) ||
-			isspace((unsigned char) str[i]) ||
-			str[i] == '\\')
+		if (strchr(NON_SIMPLE_CHARS, str[i]))
 			return false;
 	}
 	return true;
 }
 
+/*
+ * Concat multiple keys into one senna query.
+ * For senna index, "col %% 'e1' AND col %% 'e2'" will be 'e1+e2'.
+ * For like index, "col LIKE '%w1%' AND col LIKE '%w2%'" will be '"w1"+"w2"'.
+ */
 static void
-SennaScanOpen(SennaScanDesc *desc, IndexScanDesc scan, StrategyNumber strategy, Datum datum)
+SennaScanOpen(SennaScanDesc *desc,
+			  Relation indexRelation,
+			  int nkeys,
+			  const ScanKeyData keys[/*nkeys*/])
 {
+	StringInfoData	buf;
 	text		   *query;
 	const char	   *str;
 	int				len;
+	int				n;
 
-	query = DatumGetTextPP(datum);
-	str = VARDATA_ANY(query);
-	len = VARSIZE_ANY_EXHDR(query);
-
-	/* queyr is LIKE. convert LIKE query to senna query */
-	if (STRATEGY_IS_LIKE(strategy))
+	/* fast path for scanning senna index with just one key. */
+	if (nkeys == 1 && !STRATEGY_IS_LIKE(keys[0].sk_strategy))
 	{
-		if (is_simple_like(str, len))
+		query = DatumGetTextPP(keys[0].sk_argument);
+		str = VARDATA_ANY(query);
+		len = VARSIZE_ANY_EXHDR(query);
+
+		desc->records = SennaRecordsOpen(indexRelation, str, len, NULL);
+		return;
+	}
+
+	/* scan with multiple keys or LIKE strategy */
+	initStringInfo(&buf);
+	for (n = 0; n < nkeys; n++)
+	{
+		query = DatumGetTextPP(keys[n].sk_argument);
+		str = VARDATA_ANY(query);
+		len = VARSIZE_ANY_EXHDR(query);
+
+		if (STRATEGY_IS_LIKE(keys[n].sk_strategy))
 		{
-			/* don't need recheck if simple %foo% query */
-			str += 1;
-			len -= 2;
+			/* convert LIKE query to senna query */
+			if (is_simple_like(str, len))
+			{
+				/* %foo% is converted to "foo" */
+				if (buf.len > 0)
+					appendStringInfoChar(&buf, '+');
+				appendStringInfoChar(&buf, '"');
+				appendBinaryStringInfo(&buf, str + 1, len - 2);
+				appendStringInfoChar(&buf, '"');
+
+				/* don't need recheck if simple %foo% query */
+			}
+			else
+			{
+				int			i;
+				bool		quoted = false;
+
+				/* LIKE query needs recheck */
+				desc->recheck = true;
+
+				/* skip wildcard at head */
+				i = (int) strspn(str, LIKE_WILDCARD);
+				str += i;
+				len -= i;
+
+				/* convert LIKE pattern into senna query */
+				for (i = 0; i < len; i++)
+				{
+					if (strchr(WORD_SEP_CHARS, str[i]))
+					{
+						/* close the previous quote if open */
+						if (quoted)
+						{
+							appendStringInfoChar(&buf, '"');
+							quoted = false;
+						}
+					}
+					else
+					{
+						if (str[i] == '\\')
+						{
+							if (i >= len - 1)
+								ereport(ERROR,
+									(errcode(ERRCODE_INVALID_ESCAPE_SEQUENCE),
+									 errmsg("LIKE pattern must not end with escape character")));
+							/* skip the escape sequence and copy the next char. */
+							i++;
+						}
+
+						if (!quoted)
+						{
+							if (buf.len > 0)
+								appendStringInfoChar(&buf, '+');
+							appendStringInfoChar(&buf, '"');
+							quoted = true;
+						}
+
+						/* escape backslash in senna query */
+						if (str[i] == '\\')
+							appendStringInfoChar(&buf, '\\');
+						appendStringInfoChar(&buf, str[i]);
+					}
+				}
+
+				if (quoted)
+					appendStringInfoChar(&buf, '"');
+			}
 		}
 		else
 		{
-			char	   *buffer;
-			char	   *s;
-			int			i;
-			int			wilds;
-
-			/* LIKE query needs recheck */
-			desc->recheck = true;
-
-			buffer = malloc(len + LIKE_PRAGMA_LEN);
-			memcpy(buffer, LIKE_PRAGMA, LIKE_PRAGMA_LEN);
-
-			/* skip wildcard at head */
-			while (IS_LIKE_WILDCARD(*str))
-			{
-				str++;
-				len--;
-			}
-
-			/* copy query. wildcards are converted to '+' */
-			for (i = 0, s = buffer + LIKE_PRAGMA_LEN, wilds = 0; i < len; i++)
-			{
-				if (str[i] == '\\')
-				{
-					if (i >= len - 1)
-						ereport(ERROR,
-							(errcode(ERRCODE_INVALID_ESCAPE_SEQUENCE),
-							 errmsg("LIKE pattern must not end with escape character")));
-					/* skip the escape sequence and copy the next char. */
-					*s++ = str[++i];
-					wilds = 0;
-				}
-				else if (IS_LIKE_WILDCARD(str[i]))
-				{
-					if (wilds == 0)
-						*s++ = ' ';
-					wilds++;
-				}
-				else
-				{
-					*s++ = str[i];
-					wilds = 0;
-				}
-			}
-			str = buffer;
-			len = s - buffer - wilds;
+			if (buf.len > 0)
+				appendStringInfoChar(&buf, '+');
+			/* just copy for senna query */
+			appendBinaryStringInfo(&buf, str, len);
 		}
 	}
 
-	desc->records = SennaRecordsOpen(scan->indexRelation, str, len, NULL);
+	desc->records = SennaRecordsOpen(indexRelation, buf.data, buf.len, NULL);
+
+	pfree(buf.data);
 }
 
 /*
@@ -1453,13 +1511,15 @@ SennaScanNext(SennaScanDesc *desc, IndexScanDesc scan, ItemPointer ctid)
 {
 	if (desc->records == NULL)
 	{
-		/* NULL key is not supported */
-		if (scan->keyData[0].sk_flags & SK_ISNULL)
-			return false;
+		int					i,
+							nkeys = scan->numberOfKeys;
 
-		SennaScanOpen(desc, scan,
-					  scan->keyData[0].sk_strategy,
-					  scan->keyData[0].sk_argument);
+		/* NULL key is not supported */
+		for (i = 0; i < nkeys; i++)
+			if (scan->keyData[i].sk_flags & SK_ISNULL)
+				return false;
+
+		SennaScanOpen(desc, scan->indexRelation, nkeys, scan->keyData);
 	}
 
 	Assert(desc->records != NULL);
